import nexusformat.nexus as nexus

# Only in nexusformat >= 1.0.0
if not hasattr(nexus, 'NXoff_geometry'):
    NXoff_geometry = nexus.tree._makeclass('NXoff_geometry')
else:
    from nexusformat.nexus import NXoff_geometry


class NXoff():
    # Class read/create NXoff_geometry fields using an Object File Format (OFF) syntax
    def __init__(self, vertices, faces):
        self.vertices = vertices
        self.faces = faces

    @classmethod
    def from_nexus(cls, nxfield):
        # Create an OFF structure from a NeXus field
        wo = nxfield.winding_order
        fa = list(nxfield.faces) + [len(wo)]
        faces = [wo[fa[ii]:fa[ii+1]] for ii in range(len(fa)-1)]
        return cls(nxfield.vertices, faces)

    @classmethod
    def from_wedge(cls, l, w1, h1, w2=None, h2=None):
        # Create an OFF structure in shape of a wedge (trapezoidal prism)
        if w2 is None:
            w2 = w1
        if h2 is None:
            h2 = h1
        (x1, y1, x2, y2) = tuple([float(v)/2 for v in [w1, h1, w2, h2]])
        # Sets the origin at the centre of the guide entry square face
        vertices = [[-x1, -y1, 0], [-x1, y1, 0], [x1, y1, 0], [x1, -y1, 0],
                    [-x2, -y2, l], [-x2, y2, l], [x2, y2, l], [x2, -y2, l]]
        # Use a clockwise winding, facing out, beam is in +z direction
        faces = [[0, 1, 2, 3], [1, 5, 6, 2], [5, 4, 7, 6],
                 [6, 7, 3, 2], [7, 4, 0, 3], [1, 0, 4, 5]]
        return cls(vertices, faces)

    @classmethod
    def sphere(cls, radius):
        from numpy import sqrt
        phi = (1 + sqrt(5)) / 2
        r = radius / sqrt(1 + phi)
        p = r * phi
        vertices = [[-r, p, 0], [r, p, 0], [-r, -p, 0], [r, -p, 0],
                    [0, -r, p], [0, r, p], [0, -r, -p], [0, r, -p],
                    [p, 0, -r], [p, 0, r], [-p, 0, -r], [-p, 0, r]]
        faces = [[0, 11, 5], [0, 5, 1], [0, 1, 7], [0, 7, 10], [0, 10, 11],
                 [1, 5, 9], [5, 11, 4], [11, 10, 2], [10, 7, 6], [7, 1, 8],
                 [3, 9, 4], [3, 4, 2], [3, 2, 6], [3, 6, 8], [3, 8, 9],
                 [4, 9, 5], [2, 4, 11], [6, 2, 10], [8, 6, 7], [9, 8, 1]]
        return cls(vertices, faces)

    def to_nexus(self):
        from numpy import cumsum, array
        from nexusformat.nexus import NXfield
        winding_order = [ww for fa in self.faces for ww in fa]
        faces = [0] + cumsum([len(self.faces[ii]) for ii in range(len(self.faces)-1)]).tolist()
        vertices = NXfield(array(self.vertices, dtype='float64'), units='m')
        return NXoff_geometry(vertices=vertices, winding_order=winding_order, faces=faces)

    @staticmethod
    def _get_width_height(pos):
        from numpy import max, min
        # Gets the width and height (in x and y) of a set of points
        return max(pos[:, 0]) - min(pos[:, 0]), max(pos[:, 1]) - min(pos[:, 1])

    def get_guide_params(self):
        # Gets the guide parameters from the OFF geometry.
        from numpy import array, mean, where, max, min
        ve = array(self.vertices)
        zmean = mean(ve[:, 2])
        w1, h1 = self._get_width_height(ve[where(ve[:, 2] < zmean)[0], :])
        w2, h2 = self._get_width_height(ve[where(ve[:, 2] >= zmean)[0], :])
        l = max(ve[:, 2]) - min(ve[:, 2])
        return w1, h1, w2, h2, l



from networkx.classes import DiGraph
from zenlog import log
from dataclasses import dataclass, field
from mccode_antlr.instr import Instr, Instance
from mccode_antlr.instr.orientation import Orient
from mccode_antlr.common import Expr
from nexusformat.nexus import NXfield, NXgroup, NXcollection, NXinstrument
from typing import Union, Any


@dataclass
class NXInstr:
    instr: Instr
    declared: dict[str, Expr] = field(default_factory=dict)
    nxlog_root: str = field(default_factory=str)
    origin_name: Union[str, None] = None
    origin: Union[Orient, None] = None
    nx: Union[NXinstrument, None] = None
    only_nx: bool = field(default=False)
    forward_graph: Union[DiGraph, None] = None
    reverse_graph: Union[DiGraph, None] = None

    def __post_init__(self):
        """Start the C translation to ensure McCode-oddities are handled before any C-code parsing."""
        from mccode_antlr.common import ShapeType, DataType
        from mccode_antlr import Flavor
        from mccode_antlr.translators.c import CTargetVisitor
        from mccode_antlr.translators.c_listener import CDeclarator
        from mccode_antlr.translators.c_listener import evaluate_c_defined_expressions
        config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False, output=None)
        translator = CTargetVisitor(self.instr, flavor=Flavor.MCSTAS, config=config)
        # translator.instrument_uservars is a list of `CDeclaration` objects, which are named tuples with
        # fields: name type init is_pointer is_array orig
        # translator.component_uservars is a dictionary of lists for each component type of `CDeclaration` objects.

        # only worry about instrument level variables for the moment, and convert the CDeclarations into Expr objects
        def c_declaration_to_expr(dec: CDeclarator) -> Expr:
            expr = Expr._null() if dec.init is None else Expr.parse(dec.init)
            expr.data_type = DataType.from_name(dec.dtype)
            if dec.is_pointer or dec.is_array:
                expr.shape_type = ShapeType.vector
            return expr

        variables = {dec.name: c_declaration_to_expr(dec) for dec in translator.instrument_uservars}

        # defined as
        # TODO this does not work because the simple "C"-style expression parser doesn't know about pointers
        # Hopefully any %include style lines have been removed at this point.
        all_inits = '\n'.join(init.source for init in self.instr.initialize)
        try:
            variables = evaluate_c_defined_expressions(variables, all_inits)
        except AttributeError:
            log.warn(f'Evaluating INITIALIZE %{{\n{all_inits}%}}\n failed; see preceding errors for hints why. '
                     'This is not an error condition (for now). Continuing')

        self.declared = variables
        #
        if self.origin is None:
            self.guess_origin()
            assert self.origin is not None
        if self.forward_graph is None:
            self.forward_graph = self.build_graph()
        if self.reverse_graph is None:
            self.reverse_graph = self.forward_graph.reverse(copy=True)
        #
        self.make_nx_instrument()

    def guess_origin(self):
        found = (
            (lambda x: self.origin_name == x.name)
            if self.origin_name is not None else
            (lambda x: 'samples' == x.type.category)
        )
        possible = [x for x in self.instr.components if found(x)]
        if not possible:
            msg = (
                '"sample" category components'
                if self.origin_name is None else
                f'component named {self.origin_name}'
            )
            log.warn(f'No {msg} in instrument, using ABSOLUTE positions')
        elif self.origin_name is not None and len(possible) > 1:
            log.error(
                f'{len(possible)} components named "{self.origin_name}"; using the first'
            )
        elif len(possible) > 1:
            log.warn(
                'More than one "sample" category component.'
                f' Using "{possible[0].name}" for origin name'
            )
        if possible:
            self.origin_name = possible[0].name
            self.origin = possible[0].orientation
        else:
            self.origin = Orient()

    def build_graph(self):
        graph = DiGraph()
        names = [x.name for x in self.instr.components]
        graph.add_nodes_from(names)
        # Default McCode instruments are linear
        graph.add_edges_from([(names[i], names[i+1]) for i in range(len(names)-1)])
        return graph

    def inputs(self, name):
        """Return the other end of edges ending at the named node"""
        return list(self.reverse_graph[name])

    def outputs(self, name):
        """Return the other end of edges starting at the named node"""
        return list(self.forward_graph[name])

    def make_transformations(self, inst: Instance):
        from mccode_antlr.instr.orientation import Vector, Angles, Parts
        from .orientation import NXOrient, NXParts

        def abs_ref(ref):
            return f'/entry/instrument/{ref}'

        def last_ref(refs: list[tuple[str, NXfield]]) -> str | None:
            try:
                return next(reversed(refs))[0]
            except StopIteration:
                pass

        at_vec, at_rel = inst.at_relative
        rot_vec, rot_rel = inst.rotate_relative

        any_abs = at_rel is None or rot_rel is None
        nx_ori = NXOrient(self, inst.orientation - self.origin) if any_abs else None
        if at_rel is None and rot_rel is None:
            return nx_ori.transformations(inst.name)

        trans = []
        if any_abs or at_rel != rot_rel:
            import warnings
            warnings.warn(
                "All mixed-reference-type orientations untested."
                "Only 'AT (x, y, z) ABSOLUTE ROTATE (a, b, c) REF' might work"
            )
        at_vec = Vector(*at_vec) if isinstance(at_vec, tuple) else at_vec
        rot_vec = Angles(*rot_vec) if isinstance(rot_vec, tuple) else rot_vec
        if at_rel is None:
            # Absolute position with relative rotation
            trans.extend(nx_ori.position_transformations(inst.name))
            # Get the _rotation_ of the reference to add here before any new rotation
            rel_ori = NXOrient(self, rot_rel.orientation - self.origin)
            trans.extend(rel_ori.rotation_transformations(rot_rel.name, last_ref(trans)))
            # Add the relative rotation onto the reference rotation
            rot = Parts(Parts.from_at_rotated(Vector(), rot_vec, True).stack()).reduce()
            nx_parts = NXParts(self, rot, rot)
            trans.extend(nx_parts.rotation_transformations(inst.name, last_ref(trans)))
        elif rot_rel is None:
            raise RuntimeError(
                "'AT (x, y, z) RELATIVE comp ROTATE (a, b, c) ABSOLUTE'"
                " not yet implemented"
            )
        elif at_rel != rot_rel:
            raise RuntimeError(
                "'AT (x, y, z) RELATIVE comp1 ROTATE (a, b, c) RELATIVE comp2'"
                " not yet implemented"
            )
        else:
            at_parts = Parts.from_at_rotated(at_vec, Angles(), True)
            rot_parts = Parts.from_at_rotated(Vector(), rot_vec, True)
            nx_parts = NXParts(self, at_parts, rot_parts)
            # TODO replace the absolute reference to the component by an
            #      absolute reference to _its_ `depends_on` target
            target = abs_ref(at_rel.name)
            if (depends_on := self.nx.get(at_rel.name)) is not None:
                if (t := depends_on.get('depends_on')) is not None:
                    # the relative target has a dependency, so we chain off of that:
                    if isinstance(t, NXfield):
                        # TODO what if this isn't a string?
                        t = str(t)
                    if t.startswith('/'):
                        target = t
                    elif t == '.':
                        target = None
                    else:
                        target = f'{target}/{t}'
                else:
                    # the relative target exists but has no dependency, so is absolute
                    target = None
            else:
                raise RuntimeError('transformations defined out of order')
            trans.extend(nx_parts.transformations(inst.name, target))

        return {k: v for k, v in trans}

    def make_nx_instrument(self):
        from .instance import NXInstance

        self.nx = NXinstrument()
        self.nx['name'] = NXfield(value=self.instr.name)
        self.nx['mcstas'] = self.to_nx()

        for index, inst in enumerate(self.instr.components):
            transformations = self.make_transformations(inst)
            nx_inst = NXInstance(self, inst, index, transformations, only_nx=self.only_nx)
            # add input and outputs
            if len(inputs := self.inputs(inst.name)):
                nx_inst.nx.attrs['inputs'] = inputs
            if len(outputs := self.outputs(inst.name)):
                nx_inst.nx.attrs['outputs'] = outputs
            # store the nx instance representation in NeXus
            self.nx[inst.name] = nx_inst.nx


    def to_nx(self):
        # quick and very dirty:
        return NXfield(str(self.instr))

    def expr2nx(self, expr: Union[str, Expr, Any]):
        """Intended to convert *Expr* objects to NeXus-representable objects"""
        # FIXME this is called to wrap and re-wrap the same data
        #       during translation of a component with properties. It may be worth
        #       separating the parameter and component functionality.
        from moreniius.utils import link_specifier, NotNXdict
        from nexusformat.nexus import NXlog, NXfield
        if hasattr(expr, '_value') and isinstance(getattr(expr, '_value'), NotNXdict):
            # Avoid unwrapping the non-NX dictionary at this stage since it is
            # silently converted to a string-like thing which as an __iter__ property
            return expr
        if isinstance(expr, (NXlog, NXfield)):
            # Do not decompose a value if we already wrapped it in NXlog/...
            return expr
        if not isinstance(expr, str) and hasattr(expr, '__iter__'):
            parts = [self.expr2nx(x) for x in expr]
            return tuple(parts) if isinstance(expr, tuple) else parts
        if not isinstance(expr, Expr):
            return expr

        if expr.is_constant:
            return expr.value

        evaluated = expr.evaluate(self.declared)
        if evaluated.is_constant:
            return evaluated.value

        dependencies = [par.name for par in self.instr.parameters if evaluated.depends_on(par.name)]
        if len(dependencies) == 1 and str(expr) == str(dependencies[0]):
            from moreniius.utils import linked_nxlog
            return linked_nxlog(f'{self.nxlog_root}/{dependencies[0]}')

        if len(dependencies):
            log.warn(f'The expression {expr} depends on instrument parameter(s) {dependencies}\n'
                     f'A link will be inserted for each; make sure their values are stored at {self.nxlog_root}/')
            links = {par: link_specifier(par, f'{self.nxlog_root}/{par}') for par in dependencies}
            return NXcollection(expression=str(expr), **links)

        return str(expr)

    def make_nx(self, nx_class, *args, **kwargs):
        from nexusformat.nexus import NXlog
        from moreniius.utils import NotNXdict
        nx_args = [self.expr2nx(expr) for expr in args]
        nx_kwargs = {name: self.expr2nx(expr) for name, expr in kwargs.items()}

        # logged parameters are sometimes requested as NXfields, but should be NXlogs
        want_log = nx_class == NXfield and len(nx_args) == 1
        nx_arg = nx_args[0] if want_log else None
        if want_log and isinstance(nx_arg, NXlog):
            # The NXlog returned by expr2nx doesn't have the needed attributes:
            for k, v in nx_kwargs.items():
                nx_arg.attrs[k] = v
            return nx_arg
        # Hopefully less often, a collection of links in an NXcollection
        if want_log and isinstance(nx_arg, NXcollection) and 'expression' in nx_arg:
            not_expr = [x for x in nx_arg if x != 'expression']
            if len(not_expr) == 1:
                arg = nx_arg[not_expr[0]]
                if isinstance(arg, NXfield):
                    # if this is a link, we should not add any attributes
                    # since the filewriter will ignore them
                    if hasattr(arg, '_value') and isinstance(d:=getattr(arg, '_value'), NotNXdict) and d.get('module', '') == 'link':
                        return arg
                    # We have and want an NXfield, but it might be missing attributes specified in the nx_kwargs
                    # Passing the keywords to the NXfield constructor versus this method is not identical,
                    # since some keyword arguments are reserved (and only some of which are noted)
                    #   Explicit keywords, used in the constructor:
                    #       value, name, shape, dtype, group, attrs
                    #   Keywords extracted from the kwargs dict, if present (and all controlling HDF5 file attributes?):
                    #       chunks, compression, compression_opts, fillvalue, fletcher32, maxshape, scaleoffset, shuffle
                    # For now, just assume all keywords provided here are _actually_ attributes for the NXfield
                    # which is an extension of a dict, but can *not* use the update method, since the __setitem__
                    # method is overridden to wrap inputs in NXattr objects :/
                    for k, v in nx_kwargs.items():
                        arg.attrs[k] = v
                    return arg

                # TODO make this return an nx_class once we're sure that nx_kwargs is parseable (no mccode_antlr.Expr)
                if all(x in arg for x in ('module', 'config')):
                    # This is a file-writer stream directive? So make a group
                    grp = NXgroup(entries={not_expr[0]: arg})
                    for attr, val in nx_kwargs.items():
                        grp.attrs[attr] = val
                    return grp
                print('!!')
                print(arg)
                return nx_class(arg, **nx_kwargs)
            else:
                raise RuntimeError('Not sure what I should do here')
        return nx_class(*nx_args, **nx_kwargs)



"""Tests for NeXus Structure JSON path navigation utility."""

import unittest
from moreniius.path_navigator import NexusStructureNavigator


class PathNavigatorTestCase(unittest.TestCase):
    """Test cases for NexusStructureNavigator."""
    
    def setUp(self):
        """Create a sample NeXus Structure JSON for testing."""
        self.structure = {
            'children': [
                {
                    'name': 'entry',
                    'type': 'group',
                    'attributes': [
                        {'name': 'NX_class', 'dtype': 'string', 'values': 'NXentry'}
                    ],
                    'children': [
                        {
                            'name': 'instrument',
                            'type': 'group',
                            'attributes': [
                                {'name': 'NX_class', 'dtype': 'string', 'values': 'NXinstrument'}
                            ],
                            'children': [
                                {
                                    'name': 'mon0',
                                    'type': 'group',
                                    'attributes': [
                                        {'name': 'NX_class', 'dtype': 'string', 'values': 'NXmonitor'}
                                    ],
                                    'children': [
                                        {
                                            'module': 'dataset',
                                            'config': {
                                                'name': 'data',
                                                'values': [1, 2, 3],
                                                'type': 'int32'
                                            }
                                        },
                                        {
                                            'name': 'subgroup',
                                            'type': 'group',
                                            'children': []
                                        }
                                    ]
                                },
                                {
                                    'name': 'source',
                                    'type': 'group',
                                    'attributes': [
                                        {'name': 'NX_class', 'dtype': 'string', 'values': 'NXsource'}
                                    ],
                                    'children': [
                                        {
                                            'module': 'link',
                                            'config': {
                                                'name': 'distance',
                                                'source': '/entry/instrument/mon0/data'
                                            }
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        self.nav = NexusStructureNavigator(self.structure)
    
    def test_root_access(self):
        """Test accessing the root."""
        result = self.nav['']
        # Root access returns self (Navigator)
        self.assertIsInstance(result, NexusStructureNavigator)
        self.assertEqual(result.structure, self.structure)
        
        result = self.nav['/']
        self.assertIsInstance(result, NexusStructureNavigator)
        self.assertEqual(result.structure, self.structure)
    
    def test_simple_path(self):
        """Test accessing simple paths."""
        entry = self.nav['entry']
        self.assertIsInstance(entry, NexusStructureNavigator)
        self.assertEqual(entry.structure['name'], 'entry')
        self.assertEqual(entry.structure['type'], 'group')
        
        entry_slash = self.nav['/entry']
        self.assertIsInstance(entry_slash, NexusStructureNavigator)
        self.assertEqual(entry_slash.structure, entry.structure)
    
    def test_nested_path(self):
        """Test accessing nested paths."""
        instrument = self.nav['entry/instrument']
        self.assertIsInstance(instrument, NexusStructureNavigator)
        self.assertEqual(instrument.structure['name'], 'instrument')
        self.assertEqual(instrument.structure['type'], 'group')
        
        mon0 = self.nav['/entry/instrument/mon0']
        self.assertIsInstance(mon0, NexusStructureNavigator)
        self.assertEqual(mon0.structure['name'], 'mon0')
        self.assertEqual(mon0.structure['type'], 'group')
    
    def test_dataset_access(self):
        """Test accessing datasets (module='dataset')."""
        data = self.nav['/entry/instrument/mon0/data']
        self.assertEqual(data['module'], 'dataset')
        self.assertEqual(data['config']['name'], 'data')
        self.assertEqual(data['config']['values'], [1, 2, 3])
    
    def test_link_access(self):
        """Test accessing links (module='link')."""
        distance = self.nav['/entry/instrument/source/distance']
        self.assertEqual(distance['module'], 'link')
        self.assertEqual(distance['config']['name'], 'distance')
    
    def test_subgroup_access(self):
        """Test accessing nested groups."""
        subgroup = self.nav['/entry/instrument/mon0/subgroup']
        self.assertIsInstance(subgroup, NexusStructureNavigator)
        self.assertEqual(subgroup.structure['name'], 'subgroup')
        self.assertEqual(subgroup.structure['type'], 'group')
    
    def test_nonexistent_path(self):
        """Test that accessing nonexistent paths raises KeyError."""
        with self.assertRaises(KeyError) as ctx:
            _ = self.nav['/entry/nonexistent']
        self.assertIn('nonexistent', str(ctx.exception))
        
        with self.assertRaises(KeyError):
            _ = self.nav['/entry/instrument/mon0/missing']
    
    def test_get_with_default(self):
        """Test get method with default value."""
        result = self.nav.get('/entry/instrument/mon0')
        self.assertIsInstance(result, NexusStructureNavigator)
        self.assertEqual(result.structure['name'], 'mon0')
        
        result = self.nav.get('/entry/nonexistent', 'default')
        self.assertEqual(result, 'default')
        
        result = self.nav.get('/entry/nonexistent')
        self.assertIsNone(result)
    
    def test_exists(self):
        """Test exists method."""
        self.assertTrue(self.nav.exists('/entry'))
        self.assertTrue(self.nav.exists('/entry/instrument/mon0'))
        self.assertTrue(self.nav.exists('/entry/instrument/mon0/data'))
        self.assertTrue(self.nav.exists('/entry/instrument/source/distance'))
        
        self.assertFalse(self.nav.exists('/entry/nonexistent'))
        self.assertFalse(self.nav.exists('/entry/instrument/mon0/missing'))
    
    def test_find_all(self):
        """Test finding all elements with a given name."""
        # Find 'data'
        results = self.nav.find_all('data')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['module'], 'dataset')
        
        # Find 'mon0'
        results = self.nav.find_all('mon0')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'mon0')
        
        # Find something that doesn't exist
        results = self.nav.find_all('nonexistent')
        self.assertEqual(len(results), 0)
    
    def test_get_path(self):
        """Test reverse lookup - getting path to an element."""
        mon0 = self.nav['/entry/instrument/mon0']
        # get_path expects the raw dict, not the Navigator
        path = self.nav.get_path(mon0.structure)
        self.assertEqual(path, '/entry/instrument/mon0')
        
        data = self.nav['/entry/instrument/mon0/data']
        path = self.nav.get_path(data)
        self.assertEqual(path, '/entry/instrument/mon0/data')
        
        # Test with element not in structure
        fake_element = {'name': 'fake', 'type': 'group'}
        path = self.nav.get_path(fake_element)
        self.assertIsNone(path)
    
    def test_structure_property(self):
        """Test accessing the underlying structure."""
        self.assertEqual(self.nav.structure, self.structure)
    
    def test_attribute_access(self):
        """Test accessing attributes with '@' prefix."""
        # Access NX_class attribute on entry
        nx_class = self.nav['/entry/@NX_class']
        self.assertEqual(nx_class['name'], 'NX_class')
        self.assertEqual(nx_class['values'], 'NXentry')
        
        # Access NX_class on instrument
        nx_class = self.nav['/entry/instrument/@NX_class']
        self.assertEqual(nx_class['name'], 'NX_class')
        self.assertEqual(nx_class['values'], 'NXinstrument')
        
        # Access NX_class on mon0
        nx_class = self.nav['/entry/instrument/mon0/@NX_class']
        self.assertEqual(nx_class['name'], 'NX_class')
        self.assertEqual(nx_class['values'], 'NXmonitor')
    
    def test_attribute_not_found(self):
        """Test that accessing nonexistent attributes raises KeyError."""
        with self.assertRaises(KeyError) as ctx:
            _ = self.nav['/entry/@nonexistent']
        self.assertIn('nonexistent', str(ctx.exception))
    
    def test_attribute_must_be_at_end(self):
        """Test that attributes can only be at the end of path."""
        with self.assertRaises(KeyError) as ctx:
            _ = self.nav['/entry/@NX_class/something']
        self.assertIn('must be at the end', str(ctx.exception))
    
    def test_attribute_exists(self):
        """Test exists method with attributes."""
        self.assertTrue(self.nav.exists('/entry/@NX_class'))
        self.assertTrue(self.nav.exists('/entry/instrument/@NX_class'))
        self.assertTrue(self.nav.exists('/entry/instrument/mon0/@NX_class'))
        
        self.assertFalse(self.nav.exists('/entry/@nonexistent'))
    
    def test_attribute_get(self):
        """Test get method with attributes."""
        result = self.nav.get('/entry/@NX_class')
        self.assertIsNotNone(result)
        self.assertEqual(result['name'], 'NX_class')
        
        result = self.nav.get('/entry/@nonexistent', 'default')
        self.assertEqual(result, 'default')
    
    def test_find_all_attributes(self):
        """Test finding attributes with find_all."""
        # Find without including attributes (should find nothing)
        results = self.nav.find_all('NX_class', include_attributes=False)
        self.assertEqual(len(results), 0)
        
        # Find with including attributes
        results = self.nav.find_all('NX_class', include_attributes=True)
        self.assertGreater(len(results), 0)
        # Should find NX_class on entry, instrument, mon0, source, etc.
        self.assertGreaterEqual(len(results), 3)
        for result in results:
            self.assertEqual(result['name'], 'NX_class')
            self.assertIn('values', result)
    
    def test_get_path_for_attribute(self):
        """Test reverse lookup for attributes."""
        nx_class = self.nav['/entry/instrument/mon0/@NX_class']
        path = self.nav.get_path(nx_class)
        self.assertEqual(path, '/entry/instrument/mon0/@NX_class')
    
    def test_chaining_navigation(self):
        """Test chainable navigation returns Navigator objects."""
        # Single level chaining
        entry = self.nav['entry']
        self.assertIsInstance(entry, NexusStructureNavigator)
        self.assertEqual(entry.structure['name'], 'entry')
        
        # Multi-level chaining
        mon0 = self.nav['entry']['instrument']['mon0']
        self.assertIsInstance(mon0, NexusStructureNavigator)
        self.assertEqual(mon0.structure['name'], 'mon0')
        
        # Mixed chaining and path
        instrument = self.nav['entry']['instrument']
        mon0_from_instrument = instrument['mon0']
        self.assertIsInstance(mon0_from_instrument, NexusStructureNavigator)
        self.assertEqual(mon0_from_instrument.structure['name'], 'mon0')
    
    def test_dict_method(self):
        """Test .dict() method returns underlying dictionary."""
        entry = self.nav['entry']
        self.assertIsInstance(entry, NexusStructureNavigator)
        
        entry_dict = entry.dict()
        self.assertIsInstance(entry_dict, dict)
        self.assertEqual(entry_dict['name'], 'entry')
        self.assertEqual(entry_dict['type'], 'group')
        
        # dict() should be equivalent to .structure property
        self.assertEqual(entry.dict(), entry.structure)
    
    def test_attributes_return_raw_dict(self):
        """Test that attributes return raw dicts, not Navigators."""
        nx_class = self.nav['/entry/@NX_class']
        self.assertIsInstance(nx_class, dict)
        self.assertNotIsInstance(nx_class, NexusStructureNavigator)
        self.assertEqual(nx_class['name'], 'NX_class')
    
    def test_datasets_return_raw_dict(self):
        """Test that datasets return raw dicts, not Navigators."""
        data = self.nav['/entry/instrument/mon0/data']
        self.assertIsInstance(data, dict)
        self.assertNotIsInstance(data, NexusStructureNavigator)
        self.assertEqual(data['module'], 'dataset')
    
    def test_repr(self):
        """Test __repr__ provides useful information."""
        entry = self.nav['entry']
        repr_str = repr(entry)
        self.assertIn('entry', repr_str)
        self.assertIn('NexusStructureNavigator', repr_str)


class PathNavigatorRealWorldTestCase(unittest.TestCase):
    """Test with real NeXus structures from the existing tests."""
    
    def setUp(self):
        """Create a real NeXus Structure from the test instrument."""
        from json import dumps
        from mccode_antlr.loader import parse_mcstas_instr
        from moreniius.nexus_structure import to_nexus_structure
        
        # This is from test_nexus_structure.py
        instr_text = """DEFINE INSTRUMENT chopper_spectrometer(
        ch1speed, ch2speed, ch1phase, ch2phase
        )
        TRACE
        COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
        COMPONENT source = Source_simple() AT (0, 0, 0) RELATIVE origin
        COMPONENT mon0 = TOF_monitor(restore_neutron=1) AT (0, 0, 9) RELATIVE source
        COMPONENT ch1 = DiskChopper(theta_0=170, radius=0.35, nu=ch1speed, phase=ch1phase) AT (0, 0, 10) RELATIVE source
        COMPONENT ch2 = DiskChopper(theta_0=170, radius=0.35, nu=ch2speed, phase=ch2phase) AT (0, 0, 0.1) RELATIVE ch1
        COMPONENT mon1 = TOF_monitor(restore_neutron=1) AT (0, 0, 0.1) RELATIVE ch2
        COMPONENT sample = Arm() AT (0, 0, 80) RELATIVE ch2
        END
        """
        self.instr = parse_mcstas_instr(instr_text)
        self.structure = to_nexus_structure(self.instr)
        self.nav = NexusStructureNavigator(self.structure)
    
    def test_real_structure_navigation(self):
        """Test navigating the real structure."""
        # Access entry
        entry = self.nav['/entry']
        self.assertIsInstance(entry, NexusStructureNavigator)
        self.assertEqual(entry.structure['name'], 'entry')
        self.assertEqual(entry.structure['type'], 'group')
        
        # Access instrument
        instrument = self.nav['/entry/instrument']
        self.assertIsInstance(instrument, NexusStructureNavigator)
        self.assertEqual(instrument.structure['name'], 'instrument')
        self.assertEqual(instrument.structure['type'], 'group')
        
        # Access mon0
        mon0 = self.nav['/entry/instrument/mon0']
        self.assertIsInstance(mon0, NexusStructureNavigator)
        self.assertEqual(mon0.structure['name'], 'mon0')
        self.assertEqual(mon0.structure['type'], 'group')
    
    def test_real_structure_exists(self):
        """Test path existence checks on real structure."""
        self.assertTrue(self.nav.exists('/entry'))
        self.assertTrue(self.nav.exists('/entry/instrument'))
        self.assertTrue(self.nav.exists('/entry/instrument/mon0'))
        self.assertTrue(self.nav.exists('/entry/instrument/mon1'))
        self.assertTrue(self.nav.exists('/entry/instrument/sample'))
        
        self.assertFalse(self.nav.exists('/entry/nonexistent'))
    
    def test_real_structure_find_all(self):
        """Test finding components in real structure."""
        # Find all monitors
        mon0_results = self.nav.find_all('mon0')
        self.assertGreater(len(mon0_results), 0)
        
        mon1_results = self.nav.find_all('mon1')
        self.assertGreater(len(mon1_results), 0)


if __name__ == '__main__':
    unittest.main()




def make_motorized_instrument():
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler

    inst = Assembler('inst', flavor=Flavor.MCSTAS)
    inst.parameter('double ex/"m"=0')
    inst.parameter('double phi/"degree"=0')

    inst.component('origin', 'Arm', at=(0, 0, 0))
    inst.component('source', 'Source_simple', at=[(0, 0, 0), 'origin'])
    inst.component('xpos', 'Arm', at=[('ex', 0, 0), 'source'])
    inst.component('zrot', 'Arm', at=[(0, 0, 0), 'xpos'], rotate=[(0, 0, 'phi'), 'xpos'])
    inst.component('aposrot', 'Arm', at=(1, 2, 3), rotate=(45, 55, 60))

    return inst.instrument


def json_is_nxobj(obj: dict, nxtype: str):
    if 'type' not in obj or obj['type'] != 'group':
        return False
    if 'attributes' not in obj:
        return False
    for a in obj['attributes']:
        if a['name'] == 'NX_class' and a['values'] == nxtype:
            return True
    return False


def json_is_nxlog(obj: dict):
    return json_is_nxobj(obj, 'NXlog')


def get_nxlog_link_count():
    from moreniius.utils import nxlog_data_links
    return len(nxlog_data_links('fake'))


def test_motorized_instrument():
    import moreniius
    from moreniius import NexusStructureNavigator
    
    motorized = make_motorized_instrument()
    nx = moreniius.MorEniius.from_mccode(motorized, origin='origin', only_nx=False, absolute_depends_on=True)
    assert nx is not None
    
    ns_dict = nx.to_nexus_structure()
    nav = NexusStructureNavigator(ns_dict)

    # Check entry level
    entry = nav['entry']
    assert entry.structure['name'] == 'entry'
    assert entry.structure['type'] == 'group'
    assert len(entry.structure['children']) == 1
    
    entry_nx_class = entry['@NX_class']
    assert entry_nx_class == {'name': 'NX_class', 'dtype': 'string', 'values': 'NXentry'}

    # Check instrument level
    instrument = nav['entry']['instrument']
    assert instrument.structure['name'] == 'instrument'
    assert instrument.structure['type'] == 'group'
    assert len(instrument.structure['children']) == 7
    
    instrument_nx_class = instrument['@NX_class']
    assert instrument_nx_class == {'name': 'NX_class', 'dtype': 'string', 'values': 'NXinstrument'}

    # Access components directly by name using navigator
    xpos = nav['entry']['instrument']['xpos']
    zrot = nav['entry']['instrument']['zrot']
    aposrot = nav['entry']['instrument']['aposrot']

    # deps = {
    #     'xpos_t0_x': ('/entry/instrument/source', [1, 0, 0], 'translation'),
    #     # 'zrot_t0_x': ('/entry/instrument/xpos', [1, 0, 0], 'translation'), # the empty translation is skipped
    #     'zrot_r0': ('/entry/instrument/xpos', [0, 0, 1], 'rotation'), # so the rotation depends directly on xpos
    # }

    # /entry/instrumen/source is placed absolutely, so xpos_t0_x's dependency on it
    # becomes absolute as well.
    # zrot_r0 depends on xpos, but we look into its transformations group
    deps = {
        'xpos_t0_x': ('.', [1, 0, 0], 'translation'),
        'zrot_r0': ('/entry/instrument/xpos/transformations/xpos_t0_x', [0, 0, 1], 'rotation'),
    }

    for cns_nav in (xpos, zrot, aposrot):
        cns = cns_nav.structure
        assert 'children' in cns
        assert 'transformations' in [c['name'] for c in cns['children'] if 'name' in c]
        t = [c for c in cns['children'] if 'name' in c and c['name'] == 'transformations'][0]
        assert 'children' in t
        t = t['children']
        for c in t:
            # Each child can _either_ be a dataset, with 'module' at its top level
            # Or a group, with 'name', etc. at its top level

            if 'module' in c and 'dataset' == c['module']:
                # this transformation is static, and a dataset
                assert 'dataset' == c['module']
                assert all(x in c for x in ('config', 'attributes'))
                assert all(x in c['config'] for x in ('name', 'values', 'type'))
                attrs = c['attributes']
                assert len(attrs) == 4
                assert all(all(x in a for x in ('name', 'values', 'dtype')) for a in attrs)
                assert all(a['name'] in ('vector', 'depends_on', 'transformation_type', 'units') for a in attrs)
            elif 'module' in c and 'link' == c['module']:
                assert 'config' in c
                assert all(x in c['config'] for x in ('name', 'source'))
            elif json_is_nxlog(c):
                attrs = c['attributes']
                assert len(attrs) == 5

                dep = deps[c['name']]
                d = {
                    'depends_on': dep[0],
                    'vector': dep[1],
                    'transformation_type': dep[2],
                    'units': 'm' if dep[2] == 'translation' else 'degrees',
                }
                for k, v in d.items():
                    assert sum(a['name'] == k for a in attrs) == 1
                    attr = next(a for a in attrs if a['name'] == k)
                    assert attr['values'] == v

                # The children should contain a links to the log datasets
                # ... is the order important?
                assert all('module' in cc for cc in c['children'])

                assert sum('link' == cc['module'] for cc in c['children']) == get_nxlog_link_count()
                for cc in c['children']:
                    if 'link' == cc['module']:
                        assert all(x in cc['config'] for x in ('name', 'source'))

            else:
                # this transformation is dynamic and a group
                assert all(x in c for x in ('name', 'type', 'children', 'attributes'))
                assert 'group' == c['type']
                attrs = c['attributes']
                assert len(attrs) == 5

                dep = deps[c['name']]
                d = {
                    'NX_class': 'NXgroup',
                    'depends_on': dep[0],
                    'vector': dep[1],
                    'transformation_type': dep[2],
                    'units': 'm' if dep[2] == 'translation' else 'degrees',
                }
                for k, v in d.items():
                    assert sum(a['name'] == k for a in attrs) == 1
                    attr = next(a for a in attrs if a['name'] == k)
                    assert attr['values'] == v

                # The children should contain a link to the log ... is the order important?
                # Must the number of children always be the same?
                assert all('module' in cc for cc in c['children'])
                assert sum('link' == cc['module'] for cc in c['children']) <= 1
                for cc in c['children']:
                    if 'link' == cc['module']:
                        assert all(x in cc['config'] for x in ('name', 'source'))



from zenlog import log
from dataclasses import dataclass
from typing import Union
from mccode_antlr.instr import Instance
from mccode_antlr.common import Expr
from nexusformat.nexus import NXfield
from .instr import NXInstr


COMPONENT_GROUP_TO_NEXUS = dict(Guide='NXguide', Collimator='NXcollimator')
COMPONENT_CATEGORY_TO_NEXUS = dict(sources='NXmoderator', monitors='NXdetector')
COMPONENT_TYPE_NAME_TO_NEXUS = dict(
    Arm='NXcoordinate_system',
    DiskChopper='NXdisk_chopper',
    FermiChopper='NXfermi_chopper',
    FermiChopper_ILL='NXfermi_chopper',
    Fermi_chop2a='NXfermi_chopper',
    Filter_gen='NXfilter',
    Filter_graphite='NXfilter',
    Elliptic_guide_gravity='NXguide',
    Mirror='NXmirror',
    Monochromator_flat='NXmonochromator',
    Monochromator_curved='NXmonochromator',
    Monochromator_pol='NXpolarizer',
    Pol_SF_ideal='NXflipper',
    Pol_bender='NXpolarizer',
    Pol_mirror='NXpolarizer',
    SNS_source='NXmoderator',
    SNS_source_analytic='NXmoderator',
    Source_pulsed='NXmoderator',
    Selector='NXvelocity_selector',
    V_selector='NXvelocity_selector',
    ViewModISIS='NXmoderator',
)
# Each entry here maps a NeXus component to a McStas component
# The second element is a mapping of NeXus component parameters to McStas parameters
# The third element is a mapping of NeXus component parameters to McStas position paramters
NEXUS_TO_COMPONENT = dict(
    NXdetector=['Monitor_nD', {}, ],
    NXfermi_chopper=['FermiChopper',
                     {'rotation_speed': 'nu',
                      'radius': 'radius',
                      'slit': 'w',
                      'r_slit': 'curvature',
                      'number': 'nslit',
                      'width': 'xwidth',
                      'height': 'yheight'
                      },
                     {'distance': 'set_AT'
                      },
                     ],
    NXsample=['Incoherent', {}, ],
    NXmoderator=['Moderator', {}, {'distance': 'set_AT'}, ]
)


@dataclass
class NXInstance:
    instr: NXInstr
    obj: Instance
    index: int
    transforms: dict[str, NXfield]
    only_nx: bool
    nx: Union[None, dict, NXfield] = None
    dump_mcstas: bool = False

    def parameter(self, name, default=None, dtype=None):
        """
        Pull out a named instance parameter -- if it's value is not a constant, attempt to evaluate it
        using the Instr declare and initialize sections
        """
        def type_or(value):
            if dtype is not None and value is not None:
                return dtype(value)
            return value

        par = self.obj.get_parameter(name)
        if par is None:
            log.warn(f'It appears that {self.obj.type.name} does not define the parameter {name}')
            return type_or(default)

        expr = par.value
        # log.info(f'get parameter {name} which is {par}  and expr {repr(expr)}')
        if expr.is_constant:
            return type_or(expr.value)

        # Check if the expression depends on one of the instrument parameters (and thus needs a NXlog stream)
        # First resolve any declare-variables
        evaluated = expr.evaluate(self.instr.declared)
        if evaluated.is_constant:
            return type_or(evaluated.value)
        # Then check for instrument parameter(s)
        dependencies = [par for par in self.instr.instr.parameters if evaluated.depends_on(par.name)]
        log.warn(f'The parameter {name} for component instance {self.obj.name} '
                 f' depends on an instrument parameter and has value {expr}')
        return evaluated

    def expr2nx(self, expr: Expr):
        return self.instr.expr2nx(expr)

    def nx_parameter(self, name, default=None, dtype=None):
        """Retrieve the named instance parameter and convert to a NeXus compatible value"""
        return self.expr2nx(self.parameter(name, default=default, dtype=dtype))

    def make_nx(self, nx_class, *args, **kwargs):
        return self.instr.make_nx(nx_class, *args, **kwargs)

    def __post_init__(self):
        from json import dumps
        from nexusformat.nexus import NXtransformations
        from moreniius.utils import outer_transform_dependency, mccode_component_eniius_data
        self.nx = getattr(self, self.obj.type.name, self.default_translation)()
        if self.dump_mcstas:
            self.nx['mcstas'] = dumps({'instance': str(self.obj), 'order': self.index})
        if self.transforms and len(self.transforms) == 1 and ((pair := list(self.transforms.items())[0])[1] == NXfield()):
            self.nx['depends_on'] = pair[0]
        elif self.transforms:
            self.nx['transformations'] = NXtransformations(**self.transforms)
            most_dependent = outer_transform_dependency(self.nx['transformations'])
            for name, insert in mccode_component_eniius_data(self.obj, only_nx=self.only_nx).items():
                self.nx[name] = insert
                if not hasattr(self.nx[name], 'depends_on'):
                    self.nx[name].attrs['depends_on'] = most_dependent
                most_dependent = outer_transform_dependency(self.nx['transformations'])
            # Note: the depends_on entry for NX class objects is an optional dataset
            #       in the HDF5 group. Tools like chexus can verify its presence and
            #       scippnexus only uses the dataset member of a group for this.
            self.nx['depends_on'] = f'transformations/{most_dependent}'

    def get_nx_type(self):
        if self.obj.type.name in COMPONENT_TYPE_NAME_TO_NEXUS:
            return COMPONENT_TYPE_NAME_TO_NEXUS[self.obj.type.name]
        elif self.obj.type.category in COMPONENT_CATEGORY_TO_NEXUS:
            return COMPONENT_CATEGORY_TO_NEXUS[self.obj.type.category]
        if any(self.obj.type.name.startswith(x) for x in COMPONENT_GROUP_TO_NEXUS):
            return [t for k, t in COMPONENT_GROUP_TO_NEXUS.items() if self.obj.type.name.startswith(k)][0]
        return 'NXcoordinate_system' if self.transforms else 'NXnote'

    def default_translation(self):
        import nexusformat.nexus as nexus
        nx_type = self.get_nx_type()
        nx_2_mc = NEXUS_TO_COMPONENT.get(nx_type, ({}, {}))[1]
        return getattr(nexus, nx_type)(**{n: self.nx_parameter(m) for n, m in nx_2_mc.items()})


def register_translator(name, translator):
    """After the normal __init__ method instantiates a NXInstance object, the __post_init__ method is used
    to translate its held mccode_antlr.instr.Instance object into a NeXus object. This is done by calling
    the translator function registered for the component type name.
    It is not feasible to register a translator for every possible component type, so this function can be used
    to add translators for specific component types that you, the user, need for your instrument.

    Your translator must be a function with one input, the NXInstance object, and one output, a NeXus object.
    After you have defined your translator function, you can register it with this function.

    >>> import moreniius
    >>>
    >>> def my_translator(instance):
    >>>     from nexusformat.nexus import NXguide
    >>>     return instance.make_nx(NXguide, m_value=instance.parameter('m'))
    >>>
    >>> moreniius.mccode.instance.register_translator('MyComponentType', my_translator)
    """
    setattr(NXInstance, name, translator)


def register_default_translators():
    from .comp import (slit_translator, guide_translator, collimator_linear_translator,
                       diskchopper_translator, elliptic_guide_gravity_translator,
                       monitor_translator)
    for name, translator in (('Slit', slit_translator),
                             ('Guide', guide_translator),
                             ('Guide_channeled', guide_translator),
                             ('Guide_gravity', guide_translator),
                             ('Guide_simple', guide_translator),
                             ('Guide_wavy', guide_translator),
                             ('Collimator_linear', collimator_linear_translator),
                             ('DiskChopper', diskchopper_translator),
                             ('Elliptic_guide_gravity', elliptic_guide_gravity_translator),
                             ('TOF_monitor', monitor_translator),
                             ('PSD_monitor', monitor_translator),):
        register_translator(name, translator)


register_default_translators()



"""This file holds the default translators for a subset of Component types, used by NXInstance.

They can serve as a template for adding more translators, or replacing defaults, as needed.

A translator is a function that takes a component NXInstance and returns a valid NeXus object
for the _component type_ of that instance. The NXInstance holds a reference to the NXInstr object,
which can be used to look up values of identifier parameters, and the instance mccode_antlr.instr.Instance
object, which holds the instance's parameters and component type, among other information.

For a translator

def translator(nxinstance: NXInstance):
    ...

registering it as an attribute of the NXInstance class will make it available as a translator for
all instances of that class. For example:

NXInstance.SomeComponentTypeName = translator

or

NXInstance.SomeComponentTypeName = staticmethod(translator)

or

setattr(NXInstance, 'SomeComponentTypeName', translator)

will all work, even if added to the class after instances have been created.
(But since the translation is done automatically by __post_init__, it will not be applied to existing instances.)

A very nice possible enhancement to this approach would be to register each translator as a property
of the NXInstance class; then it can be used 'transparently' without needing to explicitly call it.
"""
from zenlog import log
from mccode_antlr.common import Expr
from moreniius.utils import resolve_parameter_links


def slit_translator(nxinstance):
    """The Slit component _must_ define (xmin, xmax) _or_ xwidth, and similarly the y-named parameters"""
    from nexusformat.nexus import NXslit

    def f(name):
        return nxinstance.parameter(name, dtype=float)

    if nxinstance.obj.defines_parameter('xwidth'):
        x_gap = f('xwidth')
        x_zero = Expr.float(0)
    else:
        x_gap = f('xmax') - f('xmin')
        x_zero = f('xmax') + f('xmin')
    if nxinstance.obj.defines_parameter('ywidth'):
        y_gap = f('ywidth')
        y_zero = Expr.float(0)
    else:
        y_gap = f('ymax') - f('ymin')
        y_zero = f('ymax') + f('ymin')

    if isinstance(x_zero, Expr) or isinstance(y_zero, Expr):
        log.warn(f'{nxinstance.obj.name} has a non-constant x or y zero, which requires special handling for NeXus')
    elif abs(x_zero) or abs(y_zero):
        log.warn(f'{nxinstance.obj.name} should be translated by [{x_zero}, {y_zero}, 0] via eniius_data METADATA')
    params = resolve_parameter_links(dict(x_gap=x_gap, y_gap=y_gap))
    return nxinstance.make_nx(NXslit, **params)


def guide_translator(nxinstance):
    from nexusformat.nexus import NXguide
    from moreniius.nxoff import NXoff
    off_pars = {k: nxinstance.nx_parameter(k, dtype=float) for k in ('l', 'w1', 'h1', 'w2', 'h2')}
    for k in ('w', 'h'):
        off_pars[f'{k}2'] = off_pars[f'{k}1'] if off_pars[f'{k}2'] == 0 else off_pars[f'{k}2']
    guide_pars = {'m_value': nxinstance.parameter('m', dtype=float)}
    geometry = NXoff.from_wedge(**off_pars).to_nexus()
    return nxinstance.make_nx(NXguide, OFF_GEOMETRY=geometry, **resolve_parameter_links(guide_pars))


def collimator_linear_translator(nxinstance):
    from nexusformat.nexus import NXcollimator
    from moreniius.nxoff import NXoff
    pars = {k: nxinstance.nx_parameter(v, dtype=float) for k, v in (
        ('l', 'length'), ('w1', 'xwidth'), ('h1', 'yheight')
    )}
    col_pars = dict(
        divergence_x=nxinstance.parameter('divergence', dtype=float),
        divergence_y=nxinstance.parameter('divergenceV', dtype=float)
    )
    return nxinstance.make_nx(
        NXcollimator,
        OFF_GEOMETRY=NXoff.from_wedge(**pars).to_nexus(),
        **resolve_parameter_links(col_pars)
    )


def diskchopper_translator(nxinstance):
    from nexusformat.nexus import NXdisk_chopper, NXfield
    names_types = (
        ('nslit', int),
        ('nu', float),
        ('radius', float),
        ('theta_0', float),
        ('phase', float),
        ('yheight', float),
    )
    m_pars = {k: nxinstance.parameter(k, dtype=d) for k, d in names_types}

    pars = {
        'slits': m_pars['nslit'],
        'rotation_speed': nxinstance.make_nx(NXfield, m_pars['nu'], units='Hz'),
        'radius': nxinstance.make_nx(NXfield, m_pars['radius'], units='m'),
        'slit_angle': nxinstance.make_nx(NXfield, m_pars['theta_0'], units='degrees'),
        'phase': nxinstance.make_nx(NXfield, m_pars['phase'], units='degrees'),
        'slit_height': nxinstance.make_nx(NXfield, m_pars['yheight'] if m_pars['yheight'] else m_pars['radius'], units='m')
    }
    nslit, delta = m_pars['nslit'], m_pars['theta_0'] / 2.0
    slit_edges = [y * 360.0 / nslit + x for y in range(int(nslit)) for x in (-delta, delta)]
    nx_slit_edges = [nxinstance.expr2nx(se) for se in slit_edges]
    return nxinstance.make_nx(NXdisk_chopper, slit_edges=NXfield(nx_slit_edges, units='degrees'), **resolve_parameter_links(pars))


def _ellipse_vertices_faces(*, major_x, minor_x, offset_x, major_y, minor_y, offset_y, l, n=10):
    """
    Create vertices and faces for an elliptical guide with given parameters.

    Parameters
    ----------
    major_x : float
        Major axis half-length in the x-direction.
    minor_x : float
        Minor axis half-length in the x-direction.
    offset_x : float
        Offset from the end of the ellipse to the guide entrance in the x-direction.
    major_y : float
        Major axis half-length in the y-direction.
    minor_y : float
        Minor axis half-length in the y-direction.
    offset_y : float
        Offset from the end of the ellipse to the guide entrance in the y-direction.
    l : float
        Length of the guide. l <= 2*major_x - offset_x and l <= 2*major_y - offset_y
    n : int, optional
        Number of segments along the length of the guide. Default is 10.
    """
    from numpy import arange, sqrt

    def ellipse_width(minor, major, at):
        return 0 if abs(at) > major else minor * sqrt(1 - (at / major) ** 2)

    rings = arange(n + 1) / n
    faces, vertices = [], []
    for x in rings:
        z = x * l
        w = ellipse_width(minor_x, major_x, offset_x - minor_x + z)
        h = ellipse_width(minor_y, major_y, offset_y - minor_y + x)

        vertices.extend([[-w, -h, z], [-w, h, z], [w, h, z], [w, -h, z]])

    # These are only the guide faces (that is, the inner faces of the sides of the guide housing)
    # The entry and exit are not guide faces and therefore are NOT represented here!
    for i in range(n):
        j0, j1, j2, j3, j4, j5, j6, j7 = [4 * i + k for k in range(8)]
        faces.extend([[j0, j1, j5, j4], [j1, j2, j6, j5], [j2, j3, j7, j6], [j3, j0, j4, j7]])

    return vertices, faces


def _ellipse_parameters_from_widths(nxinstance):
    from numpy import sqrt

    def parameters(which, w, i, o, l):
        foci = i + l + o
        offset = foci / 2 - i
        if 'mid' in which:
            minor = w / 2
            major = sqrt(foci ** 2 + minor ** 2) / 2
        else:
            t, b = (o, i) if 'entrance' in which else (i, o)
            t += l
            w /= 2
            b = sqrt(b * b + w * w / 4) + sqrt(t * t + w * w / 4)
            major = b / 2
            minor = sqrt(b * b - foci * foci) / 2
        return major, minor, offset

    pars = dict(xw='xwidth', xi='linxw', xo='loutxw', yw='yheight', yi='linyh', yo='loutyh', l='l')
    p = {k: nxinstance.parameter(v, dtype=float) for k, v in pars.items()}

    dim_at = str(nxinstance.obj.get_parameter('dimensionsAt').value)
    major_x, minor_x, offset_x = parameters(dim_at, p['xw'], p['xi'], p['xo'], p['l'])
    major_y, minor_y, offset_y = parameters(dim_at, p['yw'], p['yi'], p['yo'], p['l'])

    return {
        'major_x': major_x, 'minor_x': minor_x, 'offset_x': offset_x,
        'major_y': major_y, 'minor_y': minor_y, 'offset_y': offset_y,
        'l': p['l'],
    }


def elliptic_guide_gravity_translator(nxinstance):
    from nexusformat.nexus import NXguide
    from moreniius.nxoff import NXoff

    def f(name):
        return nxinstance.parameter(name, dtype=float)

    # names verified against component definition 2026-03-10:
    # https://github.com/mccode-dev/McCode/blob/main/mcstas-comps/optics/Elliptic_guide_gravity.comp
    bases = {'major': 'majorAxis', 'minor': 'minorAxis', 'offset': 'majorAxisoffset'}
    extensions = {'x': 'xw', 'y': 'yh'}
    names = {f'{a}_{b}': f'{f}{s}' for a, f in bases.items() for b, s in extensions.items()}

    pars = {k: f(v) if nxinstance.obj.defines_parameter(v) else None for k, v in names.items()}

    if len(undef := [x for x in pars.values() if x is None]) == 0:
        # the total length doesn't fit into the naming scheme
        pars['l'] = f('l')
    elif len(undef) < len(pars):
        msg = f'Only {len(pars)-len(undef)} of {len(pars)} likely parameters are defined'
        log.warn(f'Likely error state in elliptic_guide_gravity_translator: {msg}')

    if len(undef):
        # If there were any undefined parameters, we try and fall-back on calculating
        # them from the guide widths specified at 'entrance', 'mid', or ('exit')
        pars = _ellipse_parameters_from_widths(nxinstance)

    vertices, faces = _ellipse_vertices_faces(**pars, n=10)

    nx_vertices = [[nxinstance.expr2nx(expr) for expr in vector] for vector in vertices]
    nx_faces = [[nxinstance.expr2nx(expr) for expr in face] for face in faces]

    return NXguide(OFF_GEOMETRY=NXoff(nx_vertices, nx_faces).to_nexus())


def monitor_translator(nxinstance):
    from nexusformat.nexus import NXmonitor, NXdata
    from moreniius.nxoff import NXoff
    from moreniius.utils import NotNXdict
    from json import loads
    width = nxinstance.nx_parameter('xwidth', dtype=float)
    height = nxinstance.nx_parameter('yheight', dtype=float)
    geometry = NXoff.from_wedge(l=0.005, w1=width, h1=height)
    nx_monitor = NXmonitor(OFF_GEOMETRY=geometry.to_nexus())
    if len(nxinstance.obj.metadata):
        # look for mimetype 'application/json' and check if it is NeXus Structure data stream:
        for md in nxinstance.obj.metadata:
            if md.mimetype == 'application/json' and md.name == 'nexus_structure_stream_data':
                nx_monitor['data'] = NXdata(data=NotNXdict(loads(md.value)))

    return nx_monitor



import pytest
from nexusformat.nexus import NXguide


def make_elliptic_guide_instrument(use_explicit_ellipse_pars=False):
    """Create an instrument with an Elliptic_guide_gravity component."""
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler

    inst = Assembler('elliptic_guide_test', flavor=Flavor.MCSTAS)

    inst.component('origin', 'Arm', at=(0, 0, 0))
    inst.component('source', 'Source_simple', at=[(0, 0, 0), 'origin'])

    if use_explicit_ellipse_pars:
        # Use explicit ellipse axis parameters
        inst.component(
            'elliptic_guide', 'Elliptic_guide_gravity',
            at=[(0, 0, 1), 'source'],
            parameters={
                'l': 2.0,
                'majorAxisxw': 1.5,
                'minorAxisxw': 0.05,
                'majorAxisoffsetxw': 0.5,
                'majorAxisyh': 1.5,
                'minorAxisyh': 0.06,
                'majorAxisoffsetyh': 0.5,
            }
        )
    else:
        # Use width/height based parameters with dimensionsAt
        inst.component(
            'elliptic_guide', 'Elliptic_guide_gravity',
            at=[(0, 0, 1), 'source'],
            parameters={
                'l': 2.0,
                'xwidth': 0.06,
                'yheight': 0.08,
                'linxw': 1.0,
                'loutxw': 1.0,
                'linyh': 1.0,
                'loutyh': 1.0,
                'dimensionsAt': '"mid"',
            }
        )

    return inst.instrument


def test_elliptic_guide_gravity_translator_from_widths():
    """Test the elliptic guide translator using width/height parameters."""
    import moreniius

    instr = make_elliptic_guide_instrument(use_explicit_ellipse_pars=False)
    me = moreniius.MorEniius.from_mccode(
        instr, origin='origin', only_nx=False, absolute_depends_on=True
    )

    assert me is not None
    assert 'elliptic_guide' in me.nx

    guide = me.nx['elliptic_guide']
    assert isinstance(guide, NXguide)

    # Should have OFF_GEOMETRY
    assert 'OFF_GEOMETRY' in guide


def test_elliptic_guide_gravity_translator_explicit_pars():
    """Test the elliptic guide translator using explicit ellipse axis parameters."""
    import moreniius

    instr = make_elliptic_guide_instrument(use_explicit_ellipse_pars=True)
    me = moreniius.MorEniius.from_mccode(
        instr, origin='origin', only_nx=False, absolute_depends_on=True
    )

    assert me is not None
    assert 'elliptic_guide' in me.nx

    guide = me.nx['elliptic_guide']
    assert isinstance(guide, NXguide)

    # Should have OFF_GEOMETRY
    assert 'OFF_GEOMETRY' in guide


def test_elliptic_guide_geometry_has_vertices_and_faces():
    """Test that the generated geometry has vertices and faces."""
    import moreniius

    instr = make_elliptic_guide_instrument(use_explicit_ellipse_pars=True)
    me = moreniius.MorEniius.from_mccode(
        instr, origin='origin', only_nx=False, absolute_depends_on=True
    )

    guide = me.nx['elliptic_guide']
    geometry = guide['OFF_GEOMETRY']

    # NXoff_geometry should have vertices, faces, and winding_order
    assert 'vertices' in geometry
    assert 'faces' in geometry
    assert 'winding_order' in geometry

    # Vertices should be a 2D array (n_vertices, 3)
    vertices = geometry['vertices'].nxdata
    assert vertices.ndim == 2
    assert vertices.shape[1] == 3

    # With n=10 segments, we should have 11 rings of 4 vertices each = 44 vertices
    assert vertices.shape[0] == 44


def test_elliptic_guide_nexus_structure():
    """Test that the elliptic guide is correctly represented in NeXus structure output."""
    import moreniius
    from moreniius import NexusStructureNavigator

    instr = make_elliptic_guide_instrument(use_explicit_ellipse_pars=True)
    me = moreniius.MorEniius.from_mccode(
        instr, origin='origin', only_nx=False, absolute_depends_on=True
    )

    ns_dict = me.to_nexus_structure()
    
    # Use navigator for cleaner access
    nav = NexusStructureNavigator(ns_dict)

    # Navigate to entry and instrument
    entry = nav['entry']
    assert entry.structure['name'] == 'entry'

    instrument = nav['entry']['instrument']
    assert instrument.structure['name'] == 'instrument'

    # Access the elliptic_guide component directly by name
    guide = nav['entry']['instrument']['elliptic_guide']
    assert guide.structure['type'] == 'group'

    # Check that it has the NXguide class attribute using navigator
    nx_class_attr = guide['@NX_class']
    assert nx_class_attr['values'] == 'NXguide'


def test_ellipse_vertices_faces_function():
    """Test the _ellipse_vertices_faces helper function directly."""
    from moreniius.mccode.comp import _ellipse_vertices_faces

    major_x, minor_x, offset_x = 1.5, 0.05, 0.5
    major_y, minor_y, offset_y = 1.5, 0.06, 0.5
    l = 2.0
    n = 5

    vertices, faces = _ellipse_vertices_faces(
        major_x=major_x, minor_x=minor_x, offset_x=offset_x,
        major_y=major_y, minor_y=minor_y, offset_y=offset_y,
        l=l, n=n
    )

    # Should have (n+1) rings of 4 vertices each
    assert len(vertices) == (n + 1) * 4

    # Each ring creates 4 faces (top, bottom, left, right of guide)
    # For n segments, we have n * 4 faces
    assert len(faces) == n * 4

    # Each vertex should have 3 coordinates
    assert all(len(v) == 3 for v in vertices)

    # Each face should reference 4 vertex indices
    assert all(len(f) == 4 for f in faces)

    # z-coordinates should range from 0 to l
    z_coords = [v[2] for v in vertices]
    assert min(z_coords) == pytest.approx(0.0)
    assert max(z_coords) == pytest.approx(l)




from .moreniius import MorEniius
from .nexus_structure import load_instr, to_nexus_structure
from .path_navigator import NexusStructureNavigator

__all__ = ['MorEniius', 'load_instr', 'to_nexus_structure', 'NexusStructureNavigator']



from moreniius import NexusStructureNavigator


def build_bifrost_instr():
    from mccode_antlr import Flavor
    from mccode_antlr.assembler import Assembler
    from niess.bifrost.parameters import primary_parameters, tank_parameters
    from niess.bifrost import Tank, Primary

    primary = Primary.from_calibration(primary_parameters())
    tank = Tank.from_calibration(tank_parameters())

    bifrost = Assembler('bifrost', flavor=Flavor.MCSTAS)
    primary.to_mccode(bifrost)
    tank.to_mccode(bifrost, 'sample_origin')

    return bifrost.instrument


def get_bifrost_instr():
    from mccode_antlr.io.json import load_json, save_json
    from pathlib import Path
    file = Path(__file__).parent / 'bifrost.instr.json'
    try:
        instr = load_json(file)
    except:
        instr = build_bifrost_instr()
        save_json(instr, file)
    return instr

def test_bifrost_nexus_structure():
    from moreniius import MorEniius, NexusStructureNavigator
    
    instr = get_bifrost_instr()
    me = MorEniius.from_mccode(instr, origin='sample_origin', only_nx=False, absolute_depends_on=False)
    ns_dict = me.to_nexus_structure()
    
    # Use navigator for cleaner access
    nav = NexusStructureNavigator(ns_dict)

    # Old path: .children[0].children[0].children[12].children[2]
    # Navigate to pulse_shaping_chopper_1's rotation_speed
    entry = nav['entry']
    instrument = entry['instrument']
    
    # Get child #12 (pulse_shaping_chopper_1) and its child #2 (rotation_speed)
    # Since we don't know the exact name, fall back to index access for this specific case
    pulse_chopper = instrument.structure['children'][12]
    obj = pulse_chopper['children'][2]

    assert 'type' in obj
    assert obj['type'] == 'group'
    children = obj['children']
    for child in children:
        assert 'module' in child
        assert 'config' in child
        assert child['module'] == 'link'
        assert all(x in child['config'] for x in ('name', 'source'))
        assert child['config']['name'] == child['config']['source'].split('/')[-1]

    # The channel-5 arm is oriented 0-degrees rotated from the tank rotation angle
    # and _should_ be a NXcoordinate_system directly dependent on that NXlog'ed value
    # Old path: .children[0].children[0].children[255]


    # arm_dict = nav['entry']['instrument']['channel_5_arm'].structure
    # assert 'type' in arm_dict
    # assert arm_dict['type'] == 'group'
    # assert 'name' in arm_dict
    # assert arm_dict['name'] == 'channel_5_arm'
    
    # Could also access by name if it's unique
    channel_5_arm = nav['entry']['instrument']['channel_5_arm']
    assert channel_5_arm.structure['name'] == 'channel_5_arm'

    # Check attributes using navigator
    nx_class_attr = channel_5_arm['@NX_class']
    assert nx_class_attr['values'] == 'NXcoordinate_system'

    # Check children for depends_on
    assert 'children' in channel_5_arm.structure
    for child in channel_5_arm.structure['children']:
        if (m := child.get('module')) is not None and m == 'dataset':
            assert 'config' in child
            if (n := child['config'].get('name')) is not None and n == 'depends_on':
                # Check for the actual named group?
                assert child['config']['values'].startswith('/')


def is_typed_dataset(obj, typ):
    if 'module' not in obj or obj['module'] is None or obj['module'] != 'dataset':
        return False
    if 'config' not in obj or obj['config'] is None or 'values' not in obj['config']:
        return False
    return isinstance(obj['config']['values'], typ)

def is_typed_array_dataset(obj, typ, shape: list[int]):
    """
    Check if a NeXus Structure JSON object is a static dataset
    representing an array of the specified type and shape.

    Parameters
    ----------
    obj : dict
        The JSON object resulting from, e.g., the Navigator's getitem instance
    typ: typename
        The type of the dataset to check
    shape: list[int]
        The shape of the dataset to check -- C-ordered, [innermost, ..., outermost]
        Any dimension with 'expected' length -1 will match automatically
    """
    if not is_typed_dataset(obj, list):
        return False
    values = obj['config']['values']
    sh = []
    while hasattr(values, '__len__'):
        sh.append(len(values))
        values = values[0]
    if not isinstance(values, typ):
        return False
    found_shape = list(reversed(sh))
    if not len(found_shape) == len(shape):
        return False
    for i, n in enumerate(found_shape):
        if shape[i] == -1:
            shape[i] = n

    return found_shape == shape


def is_nxlog_group(loc):
    """
    Check if a NeXus Structure JSON object Navigator location is a NXlog group

    Parameters
    loc: NavigatorStructureNavigator
        A navigated-to location that should be an NXlog group
    """
    from moreniius import NexusStructureNavigator
    if not isinstance(loc, NexusStructureNavigator):
        return False
    if 'type' not in loc.structure or loc.structure['type'] != 'group':
        return False
    return loc['@NX_class']['values'] == 'NXlog'


def test_bifrost_choppers_have_necessary_parameters():
    from moreniius import MorEniius, NexusStructureNavigator

    instr = get_bifrost_instr()
    me = MorEniius.from_mccode(instr, origin='sample_origin', only_nx=False,
                               absolute_depends_on=False)
    ns_dict = me.to_nexus_structure()

    # Use navigator for cleaner access
    nav = NexusStructureNavigator(ns_dict)

    choppers = (
        'pulse_shaping_chopper_1', 'pulse_shaping_chopper_2',
        'frame_overlap_chopper_1', 'frame_overlap_chopper_2',
        'bandwidth_chopper_1', 'bandwidth_chopper_2',
    )
    for name in choppers:
        chopper = nav['entry/instrument'][name]
        assert chopper.structure['type'] == 'group'
        assert chopper['@NX_class']['values'] == 'NXdisk_chopper'

        # Every chopper needs to have static datasets:
        for prop, typ in (('slits', int), ('radius', float), ('slit_angle', float), ('slit_height', float)):
            assert is_typed_dataset(chopper[prop], typ)
        # the slit angles should be an even number of values (or always 2 for BIFROST)
        assert is_typed_array_dataset(chopper['slit_edges'], float, [2])

        # And the speed and phase (for McStas simulations) should be NXlogs
        for prop in ('rotation_speed', 'phase'):
            assert is_nxlog_group(chopper[prop])

    guides = (
        'nboa', #TODO: insert more guides here
    )
    for name in guides:
        guide = nav['entry/instrument'][name]
        assert guide.structure['type'] == 'group'
        assert guide['@NX_class']['values'] == 'NXguide'

        # TODO: (Maybe?) The NXguide constructed thus far does not include any m values
        # for prop, typ in (('m', float),):
        #     assert is_typed_dataset(guide[prop], typ) # this should be an array?

        geometry = guide['OFF_GEOMETRY']
        assert geometry.structure['type'] == 'group'
        assert geometry['@NX_class']['values'] == 'NXoff_geometry'
        assert is_typed_array_dataset(geometry['vertices'], float, [3, -1])
        assert is_typed_array_dataset(geometry['winding_order'], int, [-1])
        assert is_typed_array_dataset(geometry['faces'], int, [-1])

        vertices = geometry['vertices']['config']['values']
        winding = geometry['winding_order']['config']['values']
        faces = geometry['faces']['config']['values']

        assert 0 <= min(winding) and max(winding) < len(vertices)
        assert 0 <= min(faces) and max(faces) < len(winding)

        # If the second moment of the vertices is non-zero, the shape likely has volume
        x_avg, y_avg, z_avg = 0, 0, 0
        for x, y, z in vertices:
            x_avg += x
            y_avg += y
            z_avg += z
        x_avg /= len(vertices)
        y_avg /= len(vertices)
        z_avg /= len(vertices)

        x_sec, y_sec, z_sec = 0, 0, 0
        for x, y, z in vertices:
            x_sec += (x - x_avg)**2
            y_sec += (y - y_avg)**2
            z_sec += (z - z_avg)**2
        x_sec /= len(vertices)
        y_sec /= len(vertices)
        z_sec /= len(vertices)
        assert x_sec > 0
        assert y_sec > 0
        assert z_sec > 0





"""Example usage of NexusStructureNavigator for path-based access to NeXus Structure JSON."""

from mccode_antlr.loader import parse_mcstas_instr
from moreniius import to_nexus_structure, NexusStructureNavigator

# Define a simple instrument
instr_text = """DEFINE INSTRUMENT simple_spectrometer(param1, param2)
TRACE
COMPONENT origin = Arm() AT (0, 0, 0) ABSOLUTE
COMPONENT source = Source_simple() AT (0, 0, 0) RELATIVE origin
COMPONENT mon0 = TOF_monitor(restore_neutron=1) AT (0, 0, 9) RELATIVE source
COMPONENT sample = Arm() AT (0, 0, 80) RELATIVE mon0
END
"""

# Parse and convert to NeXus Structure JSON
instr = parse_mcstas_instr(instr_text)
structure = to_nexus_structure(instr)

# Without NexusStructureNavigator (verbose):
print("Without navigator:")
mon0_verbose = structure['children'][0]['children'][0]['children'][2]
print(f"  mon0 name: {mon0_verbose['name']}")

# With NexusStructureNavigator (clean path-based access):
print("\nWith navigator - path-based access:")
nav = NexusStructureNavigator(structure)

# Path access returns Navigator objects for groups
mon0 = nav['/entry/instrument/mon0']
print(f"  mon0 is Navigator: {isinstance(mon0, NexusStructureNavigator)}")
print(f"  mon0 name: {mon0.structure['name']}")

# Chainable navigation!
print("\nChainable navigation:")
mon0_chained = nav['entry']['instrument']['mon0']
print(f"  nav['entry']['instrument']['mon0'] works!")
print(f"  mon0 name: {mon0_chained.structure['name']}")

# Get the underlying dict with .dict() or .structure
print(f"  Using .dict(): {mon0_chained.dict()['name']}")
print(f"  Using .structure: {mon0_chained.structure['name']}")

# Access attributes with '@' prefix (returns raw dict, not Navigator)
print("\nAttribute access:")
nx_class = nav['/entry/@NX_class']
print(f"  nx_class is dict: {isinstance(nx_class, dict)}")
print(f"  Entry NX_class: {nx_class['values']}")

nx_class = nav['/entry/instrument/mon0/@NX_class']
print(f"  mon0 NX_class: {nx_class['values']}")

# Can also access attributes from Navigator objects
nx_class_from_nav = mon0['@NX_class']
print(f"  mon0['@NX_class']: {nx_class_from_nav['values']}")

# Check if paths exist
print(f"\n  Path exists /entry/instrument/mon0: {nav.exists('/entry/instrument/mon0')}")
print(f"  Path exists /entry/instrument/missing: {nav.exists('/entry/instrument/missing')}")
print(f"  Attribute exists /entry/@NX_class: {nav.exists('/entry/@NX_class')}")

# Get with default value
result = nav.get('/entry/instrument/nonexistent', default={'fallback': True})
print(f"\n  Get nonexistent path with default: {result}")

# Find all elements with a specific name
mon_results = nav.find_all('mon0')
print(f"\n  Found {len(mon_results)} element(s) named 'mon0'")

# Find all attributes with a specific name
nx_class_attrs = nav.find_all('NX_class', include_attributes=True)
print(f"  Found {len(nx_class_attrs)} NX_class attribute(s)")

# Reverse lookup - get path to an element (requires raw dict)
path = nav.get_path(mon0_chained.structure)
print(f"\n  Path to mon0 element: {path}")

# Reverse lookup for attribute
nx_class_attr = nav['/entry/@NX_class']
attr_path = nav.get_path(nx_class_attr)
print(f"  Path to NX_class attribute: {attr_path}")

# Repr is helpful
print(f"\n  Navigator repr: {repr(mon0)}")



from mccode_antlr.instr import Instance
from nexusformat.nexus import NXevent_data, NXfield, NXlog

class NotNXdict:
    """Wrapper class to prevent NXfield-parsing of the held dictionary"""
    def __init__(self, v: dict):
        self.value = v

    def to_json_dict(self):
        return self.value

    def __repr__(self):
        return f"NotNXdict<{self.value}>"

    def __str__(self):
        from json import dumps
        return dumps(self.value)

    def get(self, item, default=None):
        return self.value.get(item, default)



def outer_transform_dependency(transformations):
    """For a NXtransformations group, find the most-dependent transformation name

    E.g., for
    transforms:NXtransformations
      rotation_angle
        @depends_on=.
      chi
        @depends_on=rotation_angle
      phi
        @depends_on=chi

    find and return 'phi' since it depends on 'chi', which depends on 'rotation_angle', which is independent

    The dependency chain *must* be singular and fully contained in the NXtransformations object for this to work
    """
    names = list(transformations)
    if len(names) == 1:
        return names[0]
    def depends_on_per(name):
        obj = getattr(transformations, name)
        if not hasattr(obj, 'depends_on'):
            raise ValueError(f'{name} in {names} dependency chain missing "depends_on" attribute')
        if isinstance(obj.depends_on, NXfield):
            # obj.depends_on is an NXfield object; which has a number of properties
            #   nxgroup - the parent group
            #   dtype - string or numpy dtype
            #   shape - list or tuple of ints
            #   attrs - dict of attributes
            #   nxdata - a scalar or numpy array or string
            #   nxpath - string, where this object is in the tree
            #   nxroot - the root NXgroup object containing this object
            # I _think_ we *always* want the data stored in this object
            return obj.depends_on.nxdata
        elif isinstance(obj.depends_on, str):
            return obj.depends_on
        raise ValueError(f"depends_on attribute of {name} is not an NXfield or str")

    # depends = {name: getattr(transformations, name).depends_on for name in names}
    depends = {name: depends_on_per(name) for name in names}
    externals = [v for k, v in depends.items() if v not in depends]
    if len(externals) != 1:
        raise RuntimeError(f"Dependency chain {depends} should have one absolute dependency, found {externals} instead")

    def dep_of(name):
        d = [k for k, v in depends.items() if v == name]
        if len(d) != 1:
            raise RuntimeError(f'Expected one dependency of {name} but found dependencies: {d}')
        return d[0]

    chain = [dep_of(externals[0])]
    while len(chain) < len(names):
        chain.append(dep_of(chain[-1]))
    return chain[-1]


def _sanitize(indict):
    for k in indict:
        if isinstance(indict[k], str):
            indict[k] = indict[k].replace('<nl>', '\n').replace('<tb>', '    ')\
                .replace('<qt>', "'").replace('<bs>', '\\')
        elif isinstance(indict[k], dict):
            indict[k] = _sanitize(indict[k])
    return indict


def dict2NXobj(indict, only_nx=True):
    from nexusformat.nexus import NXfield
    import nexusformat.nexus as nexus
    outdict = {}
    def type_is_ok(t):
        return not only_nx or t.startswith('NX')

    for k, v in indict.items():
        if isinstance(v, dict) and all([f in v for f in ['type', 'value']]) and type_is_ok(v['type']):
            if v['type'] == 'NXfield':
                outdict[k] = NXfield(v['value'], **v.pop('attributes', {}))
            elif not only_nx and v['type'] == 'dict':
                outdict[k] = NotNXdict(v['value'])
            else:
                nxobj = getattr(nexus, v.pop('type'))
                outdict[k] = nxobj(**v.pop('value'))
        else:
            outdict[k] = v
    return outdict


def get_mcstasscript_component_eniius_data(comp):
    """Checks for the post-McStas 3.3 METADATA attribute, and looks there for JSON eniius_data.
    Always checks EXTEND blocks as well, for `char eniius_data[]` style entries
    """
    from re import compile

    ed = 'eniius_data'
    # METADATA is a dict[str, (str, str)] where the keys are 'names' and the tuple is (type, value)
    if hasattr(comp, 'METADATA') and ed in comp.METADATA and comp.METADATA[ed][0].lower() == 'json':
        return comp.METADATA[ed][1]
    # match any '{name}eniius_data{suffix}={encoded data};
    ed_regex = compile(rf'.*{ed}[^=]*=([^;]*);')
    ed_match = ed_regex.match(comp.EXTEND)
    if not ed_match:
        return None
    to_translate = ed_match.group(1)
    # extract the matched group, remove all \n, \, and "; replace all ' by "
    json_str = to_translate.translate(str.maketrans("'", '"', '\n"\\'))
    return json_str


def decode_component_eniius_data(comp, only_nx=True) -> dict:
    """Extract the 'eniius_data' block from the component EXTEND statement

    If found, decode the information and return a dictionary of contained NXobjects
    """
    from json import JSONDecodeError, loads
    json_str = get_mcstasscript_component_eniius_data(comp)
    if json_str is None:
        return {}
    try:
        return dict2NXobj(_sanitize(loads(json_str)), only_nx=only_nx)
    except SyntaxError:
        return {}
    except JSONDecodeError as er:
        print(f"failed to decode\n{json_str}\ndue to error {er}")
        return {}


def mccode_component_eniius_data(comp: Instance, only_nx=True) -> dict:
    """Require that any eniius data comes from a METADATA tag"""
    from json import JSONDecodeError, loads
    json_str = [md.value for md in comp.metadata if md.name == 'eniius_data' and md.mimetype.lower() == 'json']
    if not json_str:
        return {}
    if len(json_str) > 1:
        print(f'{len(json_str)} "eniius_data" METADATA blocks, using only the first one')
    try:
        return dict2NXobj(loads(json_str[0]), only_nx=only_nx)
    except JSONDecodeError as er:
        print(f'Failed to decode {json_str[0]} due to error {er}')
        return {}


def ess_flatbuffer_specifier(module: str, config: dict) -> NotNXdict:
    """ESS uses flat buffers to communicate NeXus file content information, this packages one for eniius processing"""
    return NotNXdict({'module': module, 'config': config})


def ev44_event_data_group(source: str, topic: str) -> NXevent_data:
    """Produce an NXevent_data group with ev44 stream information specified to insert group-elements

    Parameters:
        source: str
            the name of the event producer, typically matches an instrument detector name
        topic: str
            the name of the Kafka stream on which the to-be-read events are published
    """
    return NXevent_data(data=ess_flatbuffer_specifier('ev44', {'source': source, 'topic': topic}))


def nxlog_data_links(source: str):
    """
    Return link module specifications for the datasets inserted by a f144 module
    """
    datasets = (
        'alarm_message',
        'alarm_severity',
        'alarm_time',
        # 'average_value', # NeXus library failure in kafka-to-nexus:
        'connection_status',
        'connection_status_time',
        'cue_index',
        'cue_timestamp_zero',
        'description',
        # 'maximum_value', # NeXus library failure in kafka-to-nexus
        # 'minimum_value', # NeXus library failure in kafka-to-nexus
        'time',
        'value'
    )
    return {k: link_specifier(k, f'{source}/{k}') for k in datasets}


def linked_nxlog(source: str, attrs: dict | None = None) -> NXlog:
    nxlog = NXlog(**nxlog_data_links(source))
    if attrs:
        nxlog.attrs.update(attrs)
    return nxlog


def link_specifier(name: str, source: str) -> NotNXdict:
    """
    Constructs a specifier to insert a NeXus link into an ESS produced NeXus file

    Parameters:
        name: str
            the name given to the NeXus link, should be unique at its group level
        source: str
            the target of the link, the location in the eventual NeXus file where the source of data resides

    Reference:
    https://gitlab.esss.lu.se/ecdc/ess-dmsc/kafka-to-nexus/-/blob/main/documentation/commands.md?ref_type=heads#links

    Examples:
    To produce the link in the following NeXus JSON structure, one would use this function
        {
          "name": "data",
          "type": "group",
          "attributes": [
            {
              "name": "NX_class",
              "dtype": "string",
              "values": "NXdata"
            }
          ],
          "children": [
            {
              "module": "link",
              "config": {
                "name": "detector0_event_data",
                "source": "/entry/instrument/detector_panel_0/event_data"
              }
            },
          ]
        }
    one would use this function to produce an NotNXdict that the eniius writer will insert into the resulting JSON
    file _without_ converting its contents to NX* objects first.

    >>> link_specifier('detector0_event_data', '/entry/instrument/detector_panel_0/event_data')
    """
    return ess_flatbuffer_specifier('link', {'name': name, 'source': source})


def resolve_parameter_links(instance_parameters: dict):
    """Component instances have NeXus base class equivalents that require specifying any number of parameters.
    Sometimes the McCode parameters are instrument-parameters, which may only be specified at runtime.
    These component parameters are detectable and should already be replaced by NXfield objects holding NotNXdict
    objects, in turn specifying the file-writer link module that points to their real NXlog dataset.
    If the link is singular, it should have its name replaced by the NeXus class parameter so that NeXus programs
    recognize it correctly.
    """
    from nexusformat.nexus import NXfield
    for name, par in instance_parameters.items():
        if isinstance(par, NXfield) and isinstance(par.nxdata, NotNXdict) and 'link' == par.nxdata.value['module']:
            par.nxdata.value['config']['name'] = name

    return instance_parameters



"""Path-based navigation utility for NeXus Structure JSON.

NeXus Structure JSON uses a verbose hierarchical format with 'children' arrays,
making navigation cumbersome. This module provides a wrapper that enables
path-based access like: ns['/entry/instrument/component_name/field']

Attributes can be accessed with '@' prefix: ns['/entry/@NX_class']

Supports chainable navigation: ns['entry']['instrument']['component']

Quick Start:
    >>> from moreniius import to_nexus_structure, NexusStructureNavigator
    >>> structure = to_nexus_structure(instr)
    >>> nav = NexusStructureNavigator(structure)
    >>> 
    >>> # Path-based access (instead of structure['children'][0]['children'][0])
    >>> entry = nav['/entry']
    >>> instrument = nav['/entry/instrument']
    >>> component = nav['/entry/instrument/component_name']
    >>> 
    >>> # Chainable navigation (each returns a Navigator)
    >>> component = nav['entry']['instrument']['component_name']
    >>> 
    >>> # Access attributes with '@' prefix (returns raw dict)
    >>> nx_class = nav['/entry/@NX_class']
    >>> vector_attr = nav['/entry/instrument/something/@vector']
    >>> 
    >>> # Get the underlying dictionary
    >>> raw_dict = nav['entry'].dict()
    >>> # or use the property
    >>> raw_dict = nav['entry'].structure
    >>> 
    >>> # Check existence
    >>> if nav.exists('/entry/instrument/mon0'):
    ...     mon0 = nav['/entry/instrument/mon0']
    >>> 
    >>> # Safe access with default
    >>> value = nav.get('/entry/instrument/missing', default=None)
    >>> 
    >>> # Find elements by name
    >>> monitors = nav.find_all('mon0')
    >>> 
    >>> # Find attributes by name
    >>> all_nx_classes = nav.find_all('NX_class', include_attributes=True)
    >>> 
    >>> # Reverse lookup
    >>> path = nav.get_path(component.structure)  # Returns '/entry/instrument/component_name'
"""

from __future__ import annotations
from typing import Any


class NexusStructureNavigator:
    """Provides path-based navigation for NeXus Structure JSON.
    
    NeXus Structure JSON has the format:
    {
        'children': [
            {
                'name': 'entry',
                'type': 'group',
                'children': [
                    {
                        'name': 'instrument',
                        'type': 'group',
                        'children': [...]
                    }
                ]
            }
        ]
    }
    
    This class allows navigation using paths like '/entry/instrument/component'
    instead of nested dictionary access. Accessing groups returns new Navigator
    instances for chaining, while accessing attributes or leaf nodes returns
    raw dictionaries.
    
    Example:
        >>> ns = NexusStructureNavigator(nexus_structure_dict)
        >>> # Returns a new Navigator
        >>> component = ns['/entry/instrument/mon0']
        >>> # Can also chain
        >>> component = ns['entry']['instrument']['mon0']
        >>> # Get raw dictionary
        >>> raw = component.dict()
        >>> # Access attributes (returns raw dict)
        >>> nx_class = ns['/entry/instrument/mon0/@NX_class']
    """
    
    def __init__(self, structure: dict):
        """Initialize with a NeXus Structure JSON dictionary.
        
        Args:
            structure: The root NeXus Structure JSON dict (with 'children' key)
        """
        self._structure = structure
    
    def __getitem__(self, path: str) -> 'NexusStructureNavigator' | dict | Any:
        """Access elements by path.
        
        Args:
            path: A forward-slash separated path like '/entry/instrument/mon0'
                  or '/entry/instrument/mon0/@attribute_name' for attributes.
                  Leading slash is optional. Attributes are prefixed with '@'.
                  Can also be a single component name like 'entry' for chaining.
        
        Returns:
            NexusStructureNavigator if the result has children (for chaining),
            otherwise the dict/value at the specified path
            
        Raises:
            KeyError: If the path doesn't exist
        """
        if not path:
            return self
        
        # Normalize path: remove leading/trailing slashes and split
        path = path.strip('/')
        if not path:
            return self
        
        parts = path.split('/')
        result = self._navigate(self._structure, parts)
        
        # If result is a structure node (has 'children'), wrap in Navigator for chaining
        if isinstance(result, dict) and 'children' in result:
            return NexusStructureNavigator(result)
        
        # Otherwise return the raw result (attributes, datasets, etc.)
        return result
    
    def _navigate(self, current: dict, parts: list[str]) -> dict | Any:
        """Recursively navigate through the structure.
        
        Args:
            current: Current position in the structure
            parts: Remaining path parts to navigate
            
        Returns:
            The found element
            
        Raises:
            KeyError: If path doesn't exist
        """
        if not parts:
            return current
        
        target_name = parts[0]
        remaining = parts[1:]
        
        # Check if this is an attribute access (starts with '@')
        if target_name.startswith('@'):
            attr_name = target_name[1:]  # Remove '@' prefix
            if remaining:
                raise KeyError(f"Attributes must be at the end of the path: {target_name}")
            
            # Look in 'attributes' array
            attributes = current.get('attributes', [])
            for attr in attributes:
                if attr.get('name') == attr_name:
                    return attr
            
            # Attribute not found
            raise KeyError(f"Attribute '@{attr_name}' not found")
        
        # Look in 'children' array for the named element
        children = current.get('children', [])
        for child in children:
            # Match by 'name' key for groups
            if child.get('name') == target_name:
                return self._navigate(child, remaining)
            
            # For datasets (module='dataset'), check config name
            if child.get('module') == 'dataset':
                config = child.get('config', {})
                if config.get('name') == target_name:
                    return self._navigate(child, remaining)
            
            # For links (module='link'), check config name
            if child.get('module') == 'link':
                config = child.get('config', {})
                if config.get('name') == target_name:
                    return self._navigate(child, remaining)
        
        # Path not found
        path_so_far = '/'.join(parts[:len(parts) - len(remaining)])
        raise KeyError(f"Path component '{target_name}' not found at /{path_so_far}")
    
    def get(self, path: str, default: Any = None) -> dict | Any:
        """Get element by path with a default value.
        
        Args:
            path: A forward-slash separated path
            default: Value to return if path doesn't exist
            
        Returns:
            The element at the path, or default if not found
        """
        try:
            return self[path]
        except KeyError:
            return default
    
    def exists(self, path: str) -> bool:
        """Check if a path exists.
        
        Args:
            path: A forward-slash separated path
            
        Returns:
            True if the path exists, False otherwise
        """
        try:
            self[path]
            return True
        except KeyError:
            return False
    
    def find_all(self, name: str, include_attributes: bool = False) -> list[dict]:
        """Find all elements with a given name anywhere in the structure.
        
        Args:
            name: The name to search for (without '@' prefix for attributes)
            include_attributes: If True, also search in attributes arrays
            
        Returns:
            List of all matching elements
        """
        results = []
        self._find_recursive(self._structure, name, results, include_attributes)
        return results
    
    def _find_recursive(self, current: dict, name: str, results: list, include_attributes: bool = False):
        """Recursively search for elements by name."""
        # Check if current element matches
        if current.get('name') == name:
            results.append(current)
        
        # Check dataset/link config names
        if current.get('module') in ('dataset', 'link'):
            config = current.get('config', {})
            if config.get('name') == name:
                results.append(current)
        
        # Check attributes if requested
        if include_attributes:
            for attr in current.get('attributes', []):
                if attr.get('name') == name:
                    results.append(attr)
        
        # Recurse into children
        for child in current.get('children', []):
            self._find_recursive(child, name, results, include_attributes)
    
    def get_path(self, element: dict) -> str | None:
        """Get the path to an element (reverse lookup).
        
        Args:
            element: The element dict to find
            
        Returns:
            The path as a string, or None if not found
        """
        path_parts = []
        if self._build_path(self._structure, element, path_parts):
            return '/' + '/'.join(path_parts) if path_parts else '/'
        return None
    
    def _build_path(self, current: dict, target: dict, path: list[str]) -> bool:
        """Recursively build path to target element."""
        if current is target:
            return True
        
        # Check attributes
        for attr in current.get('attributes', []):
            if attr is target:
                attr_name = attr.get('name')
                if attr_name:
                    path.append(f'@{attr_name}')
                    return True
        
        # Check children
        for child in current.get('children', []):
            # Get the name of this child
            name = child.get('name')
            if not name and child.get('module') in ('dataset', 'link'):
                name = child.get('config', {}).get('name')
            
            if name:
                path.append(name)
                if self._build_path(child, target, path):
                    return True
                path.pop()
        
        return False
    
    @property
    def structure(self) -> dict:
        """Get the underlying structure dictionary."""
        return self._structure
    
    def dict(self) -> dict:
        """Get the underlying structure dictionary.
        
        This is a convenience method that's equivalent to the .structure property.
        Useful when you want the raw JSON dictionary representation.
        
        Returns:
            The underlying structure dictionary
        """
        return self._structure
    
    def __repr__(self) -> str:
        """Return a readable representation of the navigator."""
        # Get the name if this structure has one
        name = self._structure.get('name', 'root')
        node_type = self._structure.get('type', 'structure')
        return f"NexusStructureNavigator(name='{name}', type='{node_type}')"
